import sys
from pathlib import Path

path = Path(__file__).absolute().parents[1]
if str(path) not in sys.path:
    sys.path.insert(path, 0)


from jfi import jaxm

try:
    jaxm.resolve_device("cuda")
    HAS_GPU = True
except RuntimeError:
    HAS_GPU = False

def test_optax():
    try:
        import optax
    except ImportError:
        print("Module `optax` not found, skipping test")
        return

    rand_fn = lambda dtype, device: jaxm.randn(3, dtype=dtype, device=device)
    devices = [jaxm.resolve_device(x) for x in (["cpu", "cuda"] if HAS_GPU else ["cpu"])]
    for device in devices:
        for dtype in [jaxm.float32, jaxm.float64]:
            param = rand_fn(dtype, device)
            opt = optax.adam(1e-3)
            opt_state = opt.init([param])
            gs = rand_fn(dtype, device)
            updates, opt_state = opt.update([gs], opt_state, [param])
            param = optax.apply_updates([param], updates)
            print(param)


import re
from typing import Tuple, Any, Optional
import os
import copyreg

import numpy as np

from . import globals

ModuleType = type(os)  # a Python version agnostic to the get the module type


def make_random_keys(n):
    keys = globals.jrandom.split(globals.key, n + 1)
    globals.key = keys[-1]
    return keys[:-1]


def make_random_key():
    return make_random_keys(1)[0]


def copy_module(mod, recursive=True):
    new_mod = ModuleType(mod.__name__ + "_jfi")
    for attr in dir(mod):
        if not attr.startswith("_"):
            if isinstance(getattr(mod, attr), ModuleType) and recursive:
                setattr(new_mod, attr, copy_module(getattr(mod, attr), recursive=recursive))
            else:
                setattr(new_mod, attr, getattr(mod, attr))
    return new_mod


####################################################################################################


def _is_dtype(x):
    try:
        globals.jnp.dtype(x)
        return True
    except TypeError:
        return False


def resolve_device(device: Tuple[str, Any, None], idx: int = 0):
    """Convert device name to the device handle."""
    if device is None:
        #return globals.DEFAULT_DEVICE
        return None
    return globals.jax.devices(device)[idx] if isinstance(device, str) else device


def resolve_dtype(dtype):
    """Sanitize and check (via error raised) if argument is a dtype specification."""
    return globals.jnp.dtype(dtype)


def _jaxm_to(
    x: "Array",  # noqa: F821
    device_or_dtype: Optional[Tuple[str, Any]] = None,
    device: Optional[Tuple[str, Any]] = None,
    dtype: Optional[Tuple[str, Any]] = None,
):
    """Move an array to the specified device and/or dtype."""
    if device_or_dtype is not None:
        # user specifies only one argument - either a device or dtype
        try:
            dtype = resolve_dtype(device_or_dtype)
            return x.astype(dtype)
        except TypeError:
            device = resolve_device(device_or_dtype)
            return globals.jax.device_put(x, device)
    else:
        # user specifies keyword arguments for device and dtype explicitly (via kwargs)
        if device is not None and dtype is not None:
            return globals.jax.device_put(x, resolve_device(device)).astype(resolve_dtype(dtype))
        if device is not None and dtype is None:
            return globals.jax.device_put(x, resolve_device(device)).astype(
                default_dtype_for_device(device)
            )
        elif device is None and dtype is not None:
            return x.astype(resolve_dtype(dtype))
        else:
            return x


def _tree_jaxm_to(
    x: Any,
    device_or_dtype: Optional[Tuple[str, Any]] = None,
    device: Optional[Tuple[str, Any]] = None,
    dtype: Optional[Tuple[str, Any]] = None,
):
    return globals.jax.tree_util.tree_map(
        lambda z: z
        if not isinstance(z, globals.jax.Array)
        else _jaxm_to(z, device_or_dtype, device, dtype),
        x,
    )


####################################################################################################


def _make_jax_device(platform: str, id: int):
    import jax

    return jax.devices(platform)[id]


def _pickle_device(d):
    return _make_jax_device, (d.platform, d.id)


def _make_jax_array(arr_value, device):
    import jax

    return jax.device_put(arr_value, device=device)


def _pickle_array(arr):
    return _make_jax_array, (np.array(arr), arr.device())


def _enable_pickling_fixes():
    import jaxlib

    copyreg.pickle(jaxlib.xla_extension.Device, _pickle_device)
    copyreg.pickle(jaxlib.xla_extension.ArrayImpl, _pickle_array)


####################################################################################################


def default_dtype_for_device(device):
    """Convert a device to its default dtype (global dtype for CPU, float32 for GPU)."""
    device = resolve_device(device)
    if device is None or re.search("cpu", device.device_kind) is not None:
        return globals.DEFAULT_DTYPE
    else:
        return globals.jnp.float32


def get_default_dtype():
    """Returns the default dtype for CPU only, default dtype for GPU cannot be changed (float32)."""
    return globals.DEFAULT_DTYPE


def set_default_dtype(dtype):
    """Sets the default dtype for CPU only, default dtype for GPU cannot be changed (float32)."""
    globals.DEFAULT_DTYPE = resolve_dtype(dtype)


def get_default_device():
    """Returns the default device for which to create arrays on."""
    return globals.DEFAULT_DEVICE


def set_default_device(device):
    """Sets the default device for which to create arrays on."""
    globals.DEFAULT_DEVICE = resolve_device(device)
    globals.jax.config.update("jax_default_device", globals.DEFAULT_DEVICE)


def manual_seed(val):
    """Sets the random seed for the random number generator."""
    assert globals.jrandom is not None
    globals.key = globals.jrandom.PRNGKey(val)



import re

import sys
from pathlib import Path

path = Path(__file__).absolute().parents[1]
if str(path) not in sys.path:
    sys.path.insert(path, 0)

from jfi import jaxm
import jfi


def test_construction():
    jaxm.set_default_device("cpu")
    jaxm.set_default_dtype(jaxm.float64)
    # cpu first
    zs = [
        jaxm.ones(2),
        jaxm.zeros(2),
        jaxm.full(2, 2.0),
        jaxm.eye(2),
        jaxm.randn(2),
        jaxm.rand(2),
    ]
    for z in zs:
        assert z.dtype == jaxm.float64
        assert re.search("cpu", z.device().device_kind, flags=re.IGNORECASE) is not None
    # float32 on cpu next
    jaxm.set_default_dtype(jaxm.float32)
    zs = {
        "ones": jaxm.ones(2),
        "zeros": jaxm.zeros(2),
        "full": jaxm.full(2, 2.0),
        "eye": jaxm.eye(2),
        "randn": jaxm.randn(2),
        "rand": jaxm.rand(2),
    }
    for k, z in zs.items():
        msg = f"failed for {k}"
        assert z.dtype == jaxm.float32, msg
        assert re.search("cpu", z.device().device_kind, flags=re.IGNORECASE) is not None, msg
    # gpu, if available
    try:
        jaxm.jax.devices("gpu")
        has_gpu = True
    except RuntimeError:
        has_gpu = False
    if has_gpu:
        jaxm.set_default_device("gpu")
        zs = [
            jaxm.ones(2),
            jaxm.zeros(2),
            jaxm.full(2, 2.0),
            jaxm.eye(2),
            jaxm.randn(2),
            jaxm.rand(2),
        ]
        for z in zs:
            assert z.dtype == jaxm.float32
            assert re.search("cpu", z.device().device_kind, flags=re.IGNORECASE) is None
    # test if CPU allocation still works
    topts = dict(device="cpu")
    zs = [
        jaxm.ones(2, **topts),
        jaxm.zeros(2, **topts),
        jaxm.full(2, 2.0, **topts),
        jaxm.eye(2, **topts),
        jaxm.randn(2, **topts),
        jaxm.rand(2, **topts),
    ]
    for z in zs:
        assert z.dtype == jaxm.float32
        assert re.search("cpu", z.device().device_kind, flags=re.IGNORECASE) is not None
    jaxm.set_default_dtype(jaxm.float64)
    zs = [
        jaxm.ones(2, **topts),
        jaxm.zeros(2, **topts),
        jaxm.full(2, 2.0, **topts),
        jaxm.eye(2, **topts),
        jaxm.randn(2, **topts),
        jaxm.rand(2, **topts),
    ]
    for z in zs:
        assert z.dtype == jaxm.float64
        assert re.search("cpu", z.device().device_kind, flags=re.IGNORECASE) is not None
    jaxm.set_default_device("cpu")
    jaxm.set_default_dtype(jaxm.float64)


def test_moving():
    try:
        jaxm.jax.devices("gpu")
        dtypes, devices = [jaxm.float32, jaxm.float64], ["cpu", "gpu"]
    except RuntimeError:
        dtypes, devices = [jaxm.float32, jaxm.float64], ["cpu"]

    def check_dtype_device(x, dtype=None, device=None):
        if dtype is not None:
            assert x.dtype == dtype
        if device is not None:
            assert x.device() == device

    for dtype in dtypes:
        for device in devices:
            dtype, device = jfi.resolve_dtype(dtype), jfi.resolve_device(device)
            r = jaxm.randn(2, dtype=dtype, device=device)
            check_dtype_device(r, dtype=dtype, device=device)

            r = jaxm.randn(2, dtype=dtype)
            check_dtype_device(r, dtype=dtype)
            r = jaxm.randn(2, device=device)
            check_dtype_device(r, device=device)

            r = jaxm.to(jaxm.randn(2), dtype)
            check_dtype_device(r, dtype=dtype)
            r = jaxm.to(jaxm.randn(2), device)
            check_dtype_device(r, device=device)

            r = jaxm.to(jaxm.randn(2), dtype=dtype)
            check_dtype_device(r, dtype=dtype)
            r = jaxm.to(jaxm.randn(2), device=device)
            check_dtype_device(r, device=device)

            r = jaxm.to(jaxm.randn(2), dtype=dtype, device=device)
            check_dtype_device(r, dtype=dtype, device=device)


if __name__ == "__main__":
    test_construction()
    test_moving()



from .api import jaxm # noqa: F401h
from .utils import resolve_device, resolve_dtype, default_dtype_for_device # noqa: F401
from .utils import get_default_device, get_default_dtype # noqa: F401
from .utils import set_default_dtype, set_default_device # noqa: F401
from .experimental.jit import autojit, nestedautojit # noqa: F401


import sys
from pathlib import Path

path = Path(__file__).absolute().parents[1]
if str(path) not in sys.path:
    sys.path.insert(path, 0)


from jfi import jaxm


def test_randint():
    low, high, shape = -2, 5, (10, 3)
    assert jaxm.randint(low, high, shape).dtype == jaxm.int64
    assert jaxm.randint(low, high, shape, dtype=jaxm.int32).dtype == jaxm.int32


if __name__ == "__main__":
    test_randint()


import os
from subprocess import check_output
import time
import hashlib
import pickle

from . import globals
from .utils import copy_module, default_dtype_for_device, _is_dtype, resolve_device
from .utils import manual_seed, make_random_keys, make_random_key
from .utils import get_default_device, get_default_dtype, set_default_device, set_default_dtype
from .utils import _enable_pickling_fixes
from .utils import _tree_jaxm_to


def init(seed=None):
    """Initializes the wrapped jax backend setting the platform (e.g., GPU) and random seed."""

    os.environ.setdefault("JAX_ENABLE_X64", "True")
    os.environ.setdefault("XLA_PYTHON_CLIENT_PREALLOCATE", str(False))
    os.environ.setdefault("XLA_PYTHON_CLIENT_ALLOCATOR", "platform")
    if os.environ.get("JAX_PLATFORM_NAME") is None:
        try:
            check_output("nvidia-smi")
            gpu_available = True
        except FileNotFoundError:
            gpu_available = False
        os.environ["JAX_PLATFORM_NAME"] = "GPU" if gpu_available else "CPU"

    import jax
    import jax.numpy as jnp
    import jax.random as jrandom
    import jax.scipy as jsp

    globals.jax, globals.jnp, globals.jrandom, globals.jsp = jax, jnp, jrandom, jsp
    jax.config.update("jax_enable_x64", True)  # ensure float64 support

    # binding main derivatives and jit
    jaxm = copy_module(jnp, recursive=False)

    jaxm.grad, jaxm.jacobian, jaxm.hessian = jax.grad, jax.jacobian, jax.hessian
    jaxm.jvp, jaxm.vjp, jaxm.stop_gradient = jax.jvp, jax.vjp, jax.lax.stop_gradient
    jaxm.jit, jaxm.vmap = jax.jit, jax.vmap

    DEFAULT_DEVICE, DEFAULT_DTYPE = jax.devices("cpu")[0], jnp.float32
    globals.DEFAULT_DEVICE, globals.DEFAULT_DTYPE = DEFAULT_DEVICE, DEFAULT_DTYPE
    globals.jax.config.update("jax_default_device", DEFAULT_DEVICE)

    # binding random numbers
    seed = None
    seed = (
        seed
        if seed is not None
        else hash(
            hashlib.sha256(pickle.dumps((time.time(), os.getpid(), os.urandom(100)))).hexdigest()
        )
    )
    key = jrandom.PRNGKey(seed)
    globals.key = key

    def device_dtype_fn(fn, without_dtype=False, check_second_arg_for_dtype=False):
        def fn_(*args, **kw):
            device = resolve_device(kw.get("device", None))
            if "device" in kw:
                del kw["device"]
            if not without_dtype:
                kw["dtype"] = kw.get("dtype", default_dtype_for_device(device))
            if check_second_arg_for_dtype and len(args) >= 2 and _is_dtype(args[1]):
                kw["dtype"] = args[1]
                args = args[:1] + args[2:]
            return fn(*args, **kw) if device is None else jax.device_put(fn(*args, **kw), device)

        return fn_

    jaxm.array = device_dtype_fn(jnp.array, check_second_arg_for_dtype=True)
    jaxm.ones = device_dtype_fn(jnp.ones, check_second_arg_for_dtype=True)
    jaxm.zeros = device_dtype_fn(jnp.zeros, check_second_arg_for_dtype=True)
    jaxm.full = device_dtype_fn(jnp.full, check_second_arg_for_dtype=True)
    jaxm.eye = device_dtype_fn(jnp.eye)
    jaxm.arange = device_dtype_fn(jnp.arange, without_dtype=True)
    jaxm.linspace = device_dtype_fn(jnp.linspace)
    jaxm.logspace = device_dtype_fn(jnp.logspace)

    def random_fn(fn, first_arg_to_tuple=False, default_dtype=None):
        def jaxm_fn(*args, **kw):
            if "key" in kw and kw["key"] is None:
                kw = {k: v for (k, v) in kw.items() if k != "key"}
            if "key" in kw:
                key1 = kw["key"]  # use user provided key
                kw = {k: v for (k, v) in kw.items() if k != "key"}
            else:
                key1, key2 = jrandom.split(globals.key)
                if not isinstance(key2, jax.interpreters.partial_eval.DynamicJaxprTracer):
                    globals.key = key2
            # set correct device and dtype
            device = resolve_device(kw.get("device", None))
            if "device" in kw:
                del kw["device"]
            ddtype = (
                default_dtype if default_dtype is not None else default_dtype_for_device(device)
            )
            kw["dtype"] = kw.get("dtype", ddtype)
            # make the size a tuple if this is the only positional argument
            if first_arg_to_tuple and len(args) == 1 and not hasattr(args[0], "__iter__"):
                args = [(args[0],)] + list(args)[1:]
            ret = (
                fn(key1, *args, **kw)
                if device is None
                else jax.device_put(fn(key1, *args, **kw), device)
            )
            return ret

        return jaxm_fn

    jaxm.randn = random_fn(jrandom.normal, first_arg_to_tuple=True)
    jaxm.rand = random_fn(jrandom.uniform, first_arg_to_tuple=True)
    jaxm.randint = random_fn(
        lambda key, low, high, size, dtype=None: jrandom.randint(key, size, low, high, dtype=dtype),
        default_dtype=jaxm.int64,
    )

    # LA factorizations and solves
    jaxm.linalg.cholesky = jsp.linalg.cho_factor
    jaxm.linalg.cholesky_solve = jsp.linalg.cho_solve
    jaxm.linalg.lu_factor = jsp.linalg.lu_factor
    jaxm.linalg.lu_solve = jsp.linalg.lu_solve

    # some utility bindings
    jaxm.norm = jnp.linalg.norm
    jaxm.softmax = jax.nn.softmax
    jaxm.cat = jnp.concatenate
    jaxm.t = lambda x: jaxm.swapaxes(x, -1, -2)
    jaxm.nn = jax.nn
    jaxm.manual_seed = manual_seed
    jaxm.get_default_dtype = get_default_dtype
    jaxm.set_default_dtype = set_default_dtype
    jaxm.get_default_device = get_default_device
    jaxm.set_default_device = set_default_device
    jaxm.default_dtype_for_device = default_dtype_for_device
    jaxm.resolve_device = resolve_device
    jaxm.make_random_keys = make_random_keys
    jaxm.make_random_key = make_random_key
    jaxm.to = _tree_jaxm_to

    # some control flow bindings
    jaxm.cond = jax.lax.cond
    jaxm.scan = jax.lax.scan
    jaxm.while_loop = jax.lax.while_loop
    jaxm.fori_loop = jax.lax.fori_loop

    # module bindings
    jaxm.jax = jax
    jaxm.numpy = jnp
    jaxm.lax = jax.lax
    jaxm.scipy = jsp
    jaxm.random = jrandom

    _enable_pickling_fixes()
    return jaxm


####################################################################################################
jaxm = init()
globals.jaxm = jaxm



import inspect
from typing import Callable, Iterable

import numpy as np

from jax import Array, jit
from jax.tree_util import tree_flatten, tree_unflatten, tree_map


_jit_types = (Array, float, np.ndarray)
_iffy_jit_types = (int,)


class AutoJIT:
    """The AutoJIT class, which automatically compiles a function on first call,
    statically labeling non-JAX-representable arguments and keywords (e.g.,
    strings)."""

    def __init__(
        self,
        fn: Callable,
        jit_ints: bool = False,
        static_argnums: Iterable = None,
        static_argnames: Iterable = None,
    ):
        self.fn = fn
        self.compiled_fn = None
        self.fn_signature = inspect.signature(self.fn)
        self.static_argnums = set(static_argnums) if static_argnums is not None else set()
        self.static_argnames = set(static_argnames) if static_argnames is not None else set()
        if jit_ints:
            self._jit_types = _jit_types + _iffy_jit_types
        else:
            self._jit_types = _jit_types

    def __call__(self, *args, **kw):
        args_kw = self.fn_signature.bind(*args, **kw)
        args_kw.apply_defaults()
        args, kw = args_kw.args, args_kw.kwargs

        if self.compiled_fn is None:
            static_argnums = set(
                i for (i, arg) in enumerate(args) if not isinstance(arg, self._jit_types)
            )
            static_argnames = set(k for (k, v) in kw.items() if not isinstance(v, self._jit_types))

            static_argnums = static_argnums.union(self.static_argnums)
            static_argnames = static_argnames.union(self.static_argnames)

            self.compiled_fn = jit(
                self.fn,
                static_argnums=tuple(static_argnums),
                static_argnames=tuple(static_argnames),
            )
        return self.compiled_fn(*args, **kw)


autojit = AutoJIT  # alias


class NestedAutoJIT:
    """The nested version of AutoJIT, which flattens the arguments to handle pytree inputs."""

    def __init__(self, fn: Callable, jit_ints: bool = False):
        self.fn = fn
        self.compiled_fn_cache = dict()
        self.fn_signature = inspect.signature(self.fn)
        if jit_ints:
            self._jit_types = _jit_types + _iffy_jit_types
        else:
            self._jit_types = _jit_types

    def __call__(self, *args, **kw):
        args_kw = self.fn_signature.bind(*args, **kw)
        args_kw.apply_defaults()
        args, kw = args_kw.args, args_kw.kwargs

        flat_args, args_struct = tree_flatten(args)
        flat_kw, kw_struct = tree_flatten(kw)
        args_types, kw_types = tree_map(type, flat_args), tree_map(type, flat_kw)

        cache_key = (args_struct, kw_struct, tuple(args_types), tuple(kw_types))

        flat_args_kw = tuple(flat_args) + tuple(flat_kw)
        if cache_key not in self.compiled_fn_cache:

            def flat_fn(*flat_args_kw):
                args, kw_args = flat_args_kw[: len(flat_args)], flat_args_kw[len(flat_args) :]
                args = tree_unflatten(args_struct, args)
                kw = tree_unflatten(kw_struct, kw_args)
                return self.fn(*args, **kw)

            static_argnums = [
                i for (i, arg) in enumerate(flat_args_kw) if not isinstance(arg, self._jit_types)
            ]
            self.compiled_fn_cache[cache_key] = jit(flat_fn, static_argnums=static_argnums)

        return self.compiled_fn_cache[cache_key](*flat_args_kw)


nestedautojit = NestedAutoJIT  # alias



import sys
from pathlib import Path

path = Path(__file__).absolute().parents[1]
if str(path) not in sys.path:
    sys.path.insert(path, 0)


from jfi import jaxm


def fn(x, y):
    z = x * jaxm.arange(x.size, dtype=x.dtype).reshape(x.shape) - y
    z = z * jaxm.ones(z.shape, dtype=z.dtype)
    z = z + jaxm.zeros(z.shape, dtype=z.dtype)
    z = z * jaxm.linspace(0, 10, z.size, dtype=z.dtype).reshape(z.shape)
    z = z * jaxm.logspace(0, 10, z.size, dtype=z.dtype).reshape(z.shape)
    z = z + jaxm.array(x[0], dtype=x.dtype)
    return z

try:
    jaxm.resolve_device("cuda")
    HAS_GPU = True
except RuntimeError:
    HAS_GPU = False

####################################################################################################

def test_non_jit_behavior():
    devices = [jaxm.resolve_device(x) for x in (["cpu", "cuda"] if HAS_GPU else ["cpu"])]
    shape = (10, 2)

    for dtype in [jaxm.float32, jaxm.float64]:
        for device in devices:
            x = jaxm.randn(shape, dtype=dtype, device=device)
            y = jaxm.randn(shape, dtype=dtype, device=device)
            z =  fn(x, y)
            assert z.device().device_kind == device.device_kind
            assert z.dtype == dtype

def test_jit_behavior():
    devices = [jaxm.resolve_device(x) for x in (["cpu", "cuda"] if HAS_GPU else ["cpu"])]
    shape = (10, 2)

    fn_jit = jaxm.jit(fn)

    for dtype in [jaxm.float32, jaxm.float64]:
        for device in devices:
            x = jaxm.randn(shape, dtype=dtype, device=device)
            y = jaxm.randn(shape, dtype=dtype, device=device)
            z =  fn_jit(x, y)
            assert z.device().device_kind == device.device_kind
            assert z.dtype == dtype


if __name__ == "__main__":
    test_non_jit_behavior()
    test_jit_behavior()


import os
from pathlib import Path
from setuptools import setup, find_packages

setup(
    name="jfi",
    version="0.5.0",
    author="Robert Dyro",
    description="Simplified and user friendly interface to JAX, that behaves like PyTorch.",
    packages=find_packages(),
    long_description=(Path(__file__).absolute().parent / "README.md").read_text(),
)
